/*
 *
 *
 *                      JACQUET Florent -- Skia
 *                      florent.jacquet@utbm.fr
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <semaphore.h>
#include <pthread.h>
#include <time.h>

#include <utils.h>
#include <map.h>
#include <v_controller.h>

Vehicle *new_vehicle(VType t, Road *r, int g_id) {
	static int id_count = 0;
	int tmp = 0;
	Vehicle *v = malloc(sizeof(Vehicle));
	v->id = id_count++;
	v->type = t;
	v->road = r;
	if((tmp = add_to_road(r, v)) != -1)
		v->lane = tmp;
	else
		goto fail;
	v->speed = SPEED(v->type);
	v->timeout = 0;
	v->goal = g_id;
	v->offset = 0;

	return v;
fail:
	if (DEBUG) {
		fprintf(stderr, "DEBUG: Vehicle creation failed\n");
	}
	free(v);
	return NULL;
}

void free_vehicle(Vehicle *v) {
	free(v);
	if (DEBUG) fprintf(stderr, "DEBUG: Vehicle freed\n");
}

Fleet *new_fleet(int size) {
	Fleet *f = NULL;
	int i = 0;
	static int counter = 0;
	char str[7] = {'f', 'l', 'e', 'e', 't', counter%10+'0', '\0'};

	if ((f = malloc(sizeof(Fleet))) == NULL) goto fail;

	f->mut = sem_open(str, O_RDWR | O_CREAT, 0666, 1);
	if (f->mut == SEM_FAILED) {
		fprintf(stderr, "DEBUG: Semaphore error\n");
		exit(EXIT_FAILURE);
	}
	counter++;

	if((f->v = malloc(size*sizeof(Vehicle*))) == NULL) goto fail1;
	for (; i<size; i++)
		f->v[i] = NULL;
	f->size = size;
	f->nb_v = 0;

	return f;
fail1:
	sem_close(f->mut);
	free(f);
fail:
	return NULL;
}

void free_fleet(Fleet *f) {
	int i = 0;
	for (; i < f->nb_v; i++) {
		free_vehicle(f->v[i]);
	}
	sem_close(f->mut);
	free(f->v);
	free(f);
	if(DEBUG) fprintf(stderr, "DEBUG: Fleet freed\n");
}

/*\
 * This is the threaded function to handle a fleet of one type of vehicle every
 * clock cycle
\*/
void *start_controller(void *a) {
	void **arg = (void**)a;
	sem_t *s = arg[0];
	VType t = (VType) arg[1];
	int nb_v_max;
	int r_nb = 0;
	int i = 0, j = 0;
	int timeout = 1;
	Map *map = (Map*) arg[2];
	Fleet *f = NULL;
	int goals[8] = {03, 05, 11, 14, 22, 27, 30, 36};

	switch(t){
		case CAR:
			nb_v_max = 100;
			printf("CAR\n");
			break;
		case TRUCK:
			nb_v_max = 100;
			printf("TRUCK\n");
			break;
		case MOTOBIKE:
			nb_v_max = 100;
			printf("MOTO\n");
			break;
		default:
			nb_v_max = 15;
	}
	f = new_fleet(nb_v_max);
	sem_wait(s);
	if (map != NULL) {
		printf("Fleet number: %d\n", map->fleet_nb);
		map->fleets[map->fleet_nb++] = f;
	}
	sem_post(s);
	r_nb = map->road_nb;

	/* Main loop */
	while(run) {
		for (i=0;i<r_nb;i++) {
			Road *r = map->roads[i];
			sem_wait(map->road_mut[r->id]);
			if(timeout <= 0 && r->input && f->nb_v < f->size) {
				int goal = goals[(time(NULL))%8]; /* Pick one of the possible goal number */
				sem_wait(f->mut);
				f->v[f->nb_v] = new_vehicle(t, r, goal);
				if(f->v[f->nb_v] != NULL) {
					if(DEBUG)
						fprintf(stderr, "DEBUG: New vehicle created\nDEBUG: Vehicle nb: %d\n", f->nb_v);
					f->nb_v++;
				}
				sem_post(f->mut);
				timeout = SPAWN_FREQUENCY*r_nb*2+time(NULL)%8+1;
			} else {
				timeout--;
			}
			for (j=0;j<f->nb_v;j++){
				if (f->v[j]->road->id == r->id) {
					if (step_vehicle(f->v[j]) == 1) {
						sem_wait(f->mut);
						f->v[j]->road->lane[f->v[j]->lane][f->v[j]->offset] = NULL;
						free_vehicle(f->v[j]);
						f->nb_v--;
						f->v[j] = f->v[f->nb_v];
						f->v[f->nb_v] = NULL;
						j--;
						sem_post(f->mut);
					}
				}
			}
			sem_post(map->road_mut[r->id]);
		}

		usleep(CLOCK);
	}
	return NULL;
}

int step_vehicle(Vehicle *v) {
	int os = v->offset;
	int width = v->road->width;
	int max_os = v->road->length-1;
	int new_os = os+1;
	int ln = v->lane;
	if (os >= max_os) {
		if (v->road->end == NULL)
			return 1;
		return 0;
	} else {
		if(v->timeout <= 0) {
			new_os = os+1; /* First targeted offset */
			v->timeout = v->speed;
		} else {
			v->timeout--;
			return 0;
		}
	}

	if (v->road->lane[ln][new_os] == NULL) {
		v->road->lane[ln][os] = NULL;
		v->road->lane[ln][new_os] = v;
		v->offset = new_os;
	} else { /* Something is ahead */
		/* Let's try to find a free lane */
		int i = 0;
		if (DEBUG)
			fprintf(stderr, "DEBUG: %d\n", v->id);
		if (ln > 0 && v->road->lane[ln-1][new_os] == NULL) i = ln-1;
		else if (ln < width-1 && v->road->lane[ln+1][new_os] == NULL) i = ln+1;
		else return 0;

		v->road->lane[ln][os] = NULL;
		v->road->lane[i][new_os] = v;
		v->offset = new_os;
		v->lane = i;
	}
	return 0;
}


